@@ -1,11 +1,11 @@
+import asyncio
 import time
 
 import pytest
-from unittest.mock import patch, Mock
 
-from mobiletestorchestrator import Device
-from mobiletestorchestrator.application import Application
-from mobiletestorchestrator.device import DeviceInteraction
+from mobiletestorchestrator.device import Device
+from mobiletestorchestrator.application import Application, AsyncApplication
+from mobiletestorchestrator.device_interactions import DeviceInteraction, AsyncDeviceInteraction
 
 
 class TestDeviceInteraction:
@@ -24,15 +24,11 @@ class TestDeviceInteraction:
 
     def test_return_home_succeeds(self, install_app, device: Device, support_app: str):
         app = install_app(Application, support_app)
-        with patch('mobiletestorchestrator.device.DeviceInteraction.home_screen_active',
-                   new_callable=Mock) as mock_home_screen_active:
-            # Have to mock out call since inputting the KEYCODE_BACK event doesn't work for all devices/emulators
-            mock_home_screen_active.return_value = True
-            app.start(activity=".MainActivity")
-            assert device.foreground_activity() == app.package_name
-            device_nav = DeviceInteraction(device)
-            device_nav.return_home()
-            assert device_nav.home_screen_active()
+        app.start(activity=".MainActivity")
+        assert device.foreground_activity() == app.package_name
+        device_nav = DeviceInteraction(device)
+        device_nav.return_home()
+        assert device_nav.home_screen_active()
 
     def test_return_home_fails(self, install_app, device: Device, support_app: str):
         app = install_app(Application, support_app)
@@ -43,3 +39,28 @@ class TestDeviceInteraction:
             # need to pass -1 to force the function to reach its back button key-press limit
             DeviceInteraction(device).return_home(keycode_back_limit=-1)
         assert "Max number of back button presses" in str(excinfo.value)
+
+
+class TestDeviceInteractionAsync:
+
+    @pytest.mark.asyncio
+    async def test_is_screen_on(self, device: Device):
+        navigator = AsyncDeviceInteraction(device)
+        is_screen_on = navigator.is_screen_on()
+        await navigator.toggle_screen_on()
+        retries = 3
+        new_is_screen_on = is_screen_on
+        while retries > 0 and new_is_screen_on == is_screen_on:
+            await asyncio.sleep(1)
+            new_is_screen_on = await navigator.is_screen_on()
+            retries -= 1
+        assert is_screen_on != new_is_screen_on
+
+    @pytest.mark.asyncio
+    async def test_go_home(self, device: Device, install_app_async, support_app: Application):
+        app = await install_app_async(AsyncApplication, support_app)
+        await app.start(activity=".MainActivity")
+        assert device.foreground_activity() == app.package_name
+        device_nav = AsyncDeviceInteraction(device)
+        await device_nav.return_home()
+        assert device_nav.home_screen_active()
